@@ -1,5 +1,5 @@
 /*
- * Copyright 2016-2024 The OpenSSL Project Authors. All Rights Reserved.
+ * Copyright 2016-2025 The OpenSSL Project Authors. All Rights Reserved.
  *
  * Licensed under the Apache License 2.0 (the "License").  You may not use
  * this file except in compliance with the License.  You can obtain a copy
@@ -12,11 +12,11 @@
  * deprecated function.
  */
 #ifndef OPENSSL_NO_DEPRECATED_3_0
-# define OPENSSL_SUPPRESS_DEPRECATED
+#define OPENSSL_SUPPRESS_DEPRECATED
 #endif
 
 #if defined(_WIN32)
-# include <windows.h>
+#include <windows.h>
 #endif
 
 #include <string.h>
@@ -42,13 +42,14 @@
 #endif
 
 /* Limit the maximum number of threads */
-#define MAXIMUM_THREADS     10
+#define MAXIMUM_THREADS 10
 
 /* Limit the maximum number of providers loaded into a library context */
-#define MAXIMUM_PROVIDERS   4
+#define MAXIMUM_PROVIDERS 4
 
 static int do_fips = 0;
 static char *privkey;
+static char *storedir;
 static char *config_file = NULL;
 static int multidefault_run = 0;
 
@@ -90,10 +91,13 @@ static int test_lock(void)
     CRYPTO_RWLOCK *lock = CRYPTO_THREAD_lock_new();
     int res;
 
+    if (!TEST_ptr(lock))
+        return 0;
+
     res = TEST_true(CRYPTO_THREAD_read_lock(lock))
-          && TEST_true(CRYPTO_THREAD_unlock(lock))
-          && TEST_true(CRYPTO_THREAD_write_lock(lock))
-          && TEST_true(CRYPTO_THREAD_unlock(lock));
+        && TEST_true(CRYPTO_THREAD_unlock(lock))
+        && TEST_true(CRYPTO_THREAD_write_lock(lock))
+        && TEST_true(CRYPTO_THREAD_unlock(lock));
 
     CRYPTO_THREAD_lock_free(lock);
 
@@ -111,6 +115,7 @@ static int rwwriter2_iterations = 0;
 static int *rwwriter_ptr = NULL;
 static int rw_torture_result = 1;
 static CRYPTO_RWLOCK *rwtorturelock = NULL;
+static CRYPTO_RWLOCK *atomiclock = NULL;
 
 static void rwwriter_fn(int id, int *iterations)
 {
@@ -119,8 +124,8 @@ static void rwwriter_fn(int id, int *iterations)
     OSSL_TIME t1, t2;
     t1 = ossl_time_now();
 
-    for (count = 0; ; count++) {
-        new = CRYPTO_zalloc(sizeof (int), NULL, 0);
+    for (count = 0;; count++) {
+        new = CRYPTO_zalloc(sizeof(int), NULL, 0);
         if (contention == 0)
             OSSL_sleep(1000);
         if (!CRYPTO_THREAD_write_lock(rwtorturelock))
@@ -150,7 +155,7 @@ static void rwwriter1_fn(void)
 
     TEST_info("Starting writer1");
     rwwriter_fn(1, &rwwriter1_iterations);
-    CRYPTO_atomic_add(&rwwriter1_done, 1, &local, NULL);
+    CRYPTO_atomic_add(&rwwriter1_done, 1, &local, atomiclock);
 }
 
 static void rwwriter2_fn(void)
@@ -159,7 +164,7 @@ static void rwwriter2_fn(void)
 
     TEST_info("Starting writer 2");
     rwwriter_fn(2, &rwwriter2_iterations);
-    CRYPTO_atomic_add(&rwwriter2_done, 1, &local, NULL);
+    CRYPTO_atomic_add(&rwwriter2_done, 1, &local, atomiclock);
 }
 
 static void rwreader_fn(int *iterations)
@@ -171,20 +176,23 @@ static void rwreader_fn(int *iterations)
     int lw2 = 0;
 
     if (CRYPTO_THREAD_read_lock(rwtorturelock) == 0)
-            abort();
+        abort();
 
     while (lw1 != 1 || lw2 != 1) {
-        CRYPTO_atomic_add(&rwwriter1_done, 0, &lw1, NULL);
-        CRYPTO_atomic_add(&rwwriter2_done, 0, &lw2, NULL);
+        CRYPTO_atomic_add(&rwwriter1_done, 0, &lw1, atomiclock);
+        CRYPTO_atomic_add(&rwwriter2_done, 0, &lw2, atomiclock);
 
         count++;
-        if (rwwriter_ptr != NULL && old > *rwwriter_ptr) {
-            TEST_info("rwwriter pointer went backwards\n");
-            rw_torture_result = 0;
+        if (rwwriter_ptr != NULL) {
+            if (old > *rwwriter_ptr) {
+                TEST_info("rwwriter pointer went backwards! %d : %d\n",
+                    old, *rwwriter_ptr);
+                rw_torture_result = 0;
+            }
+            old = *rwwriter_ptr;
         }
         if (CRYPTO_THREAD_unlock(rwtorturelock) == 0)
             abort();
-        *iterations = count;
         if (rw_torture_result == 0) {
             *iterations = count;
             return;
@@ -194,7 +202,7 @@ static void rwreader_fn(int *iterations)
     }
     *iterations = count;
     if (CRYPTO_THREAD_unlock(rwtorturelock) == 0)
-            abort();
+        abort();
 }
 
 static void rwreader1_fn(void)
@@ -223,6 +231,10 @@ static int _torture_rw(void)
     struct timeval dtime;
 
     rwtorturelock = CRYPTO_THREAD_lock_new();
+    atomiclock = CRYPTO_THREAD_lock_new();
+    if (!TEST_ptr(rwtorturelock) || !TEST_ptr(atomiclock))
+        goto out;
+
     rwwriter1_iterations = 0;
     rwwriter2_iterations = 0;
     rwreader1_iterations = 0;
@@ -253,8 +265,13 @@ static int _torture_rw(void)
     tottime = dtime.tv_sec + (dtime.tv_usec / 1e6);
     TEST_info("rw_torture_result is %d\n", rw_torture_result);
     TEST_info("performed %d reads and %d writes over 2 read and 2 write threads in %e seconds",
-              rwreader1_iterations + rwreader2_iterations,
-              rwwriter1_iterations + rwwriter2_iterations, tottime);
+        rwreader1_iterations + rwreader2_iterations,
+        rwwriter1_iterations + rwwriter2_iterations, tottime);
+    if ((rwreader1_iterations + rwreader2_iterations == 0)
+        || (rwwriter1_iterations + rwwriter2_iterations == 0)) {
+        TEST_info("Threads did not iterate\n");
+        goto out;
+    }
     avr = tottime / (rwreader1_iterations + rwreader2_iterations);
     avw = (tottime / (rwwriter1_iterations + rwwriter2_iterations));
     TEST_info("Average read time %e/read", avr);
@@ -264,6 +281,7 @@ static int _torture_rw(void)
         ret = 1;
 out:
     CRYPTO_THREAD_lock_free(rwtorturelock);
+    CRYPTO_THREAD_lock_free(atomiclock);
     rwtorturelock = NULL;
     return ret;
 }
@@ -280,8 +298,6 @@ static int torture_rw_high(void)
     return _torture_rw();
 }
 
-
-# ifndef OPENSSL_SYS_MACOSX 
 static CRYPTO_RCU_LOCK *rcu_lock = NULL;
 
 static int writer1_done = 0;
@@ -306,8 +322,9 @@ static void writer_fn(int id, int *iterations)
 
     t1 = ossl_time_now();
 
-    for (count = 0; ; count++) {
-        new = CRYPTO_zalloc(sizeof(uint64_t), NULL, 0);
+    for (count = 0;; count++) {
+        new = CRYPTO_malloc(sizeof(uint64_t), NULL, 0);
+        *new = (uint64_t)0xBAD;
         if (contention == 0)
             OSSL_sleep(1000);
         ossl_rcu_write_lock(rcu_lock);
@@ -336,7 +353,7 @@ static void writer1_fn(void)
 
     TEST_info("Starting writer1");
     writer_fn(1, &writer1_iterations);
-    CRYPTO_atomic_add(&writer1_done, 1, &local, NULL);
+    CRYPTO_atomic_add(&writer1_done, 1, &local, atomiclock);
 }
 
 static void writer2_fn(void)
@@ -345,7 +362,7 @@ static void writer2_fn(void)
 
     TEST_info("Starting writer2");
     writer_fn(2, &writer2_iterations);
-    CRYPTO_atomic_add(&writer2_done, 1, &local, NULL);
+    CRYPTO_atomic_add(&writer2_done, 1, &local, atomiclock);
 }
 
 static void reader_fn(int *iterations)
@@ -358,15 +375,23 @@ static void reader_fn(int *iterations)
     int lw2 = 0;
 
     while (lw1 != 1 || lw2 != 1) {
-        CRYPTO_atomic_add(&writer1_done, 0, &lw1, NULL);
-        CRYPTO_atomic_add(&writer2_done, 0, &lw2, NULL);
+        CRYPTO_atomic_add(&writer1_done, 0, &lw1, atomiclock);
+        CRYPTO_atomic_add(&writer2_done, 0, &lw2, atomiclock);
         count++;
-        ossl_rcu_read_lock(rcu_lock);
+        if (!ossl_rcu_read_lock(rcu_lock)) {
+            TEST_info("rcu torture read lock failed");
+            rcu_torture_result = 0;
+            *iterations = count;
+            return;
+        }
+
         valp = ossl_rcu_deref(&writer_ptr);
         val = (valp == NULL) ? 0 : *valp;
 
         if (oldval > val) {
             TEST_info("rcu torture value went backwards! %llu : %llu", (unsigned long long)oldval, (unsigned long long)val);
+            if (valp == NULL)
+                TEST_info("ossl_rcu_deref did return NULL!");
             rcu_torture_result = 0;
         }
         oldval = val; /* just try to deref the pointer */
@@ -402,6 +427,11 @@ static int _torture_rcu(void)
     struct timeval dtime;
     double tottime;
     double avr, avw;
+    int rc = 0;
+
+    atomiclock = CRYPTO_THREAD_lock_new();
+    if (!TEST_ptr(atomiclock))
+        goto out;
 
     memset(&writer1, 0, sizeof(thread_t));
     memset(&writer2, 0, sizeof(thread_t));
@@ -416,7 +446,9 @@ static int _torture_rcu(void)
     writer2_done = 0;
     rcu_torture_result = 1;
 
-    rcu_lock = ossl_rcu_lock_new(1);
+    rcu_lock = ossl_rcu_lock_new(contention == 2 ? 4 : 1, NULL);
+    if (rcu_lock == NULL)
+        goto out;
 
     TEST_info("Staring rcu torture");
     t1 = ossl_time_now();
@@ -428,25 +460,36 @@ static int _torture_rcu(void)
         || !TEST_true(wait_for_thread(writer2))
         || !TEST_true(wait_for_thread(reader1))
         || !TEST_true(wait_for_thread(reader2)))
-        return 0;
+        goto out;
 
     t2 = ossl_time_now();
     dtime = ossl_time_to_timeval(ossl_time_subtract(t2, t1));
     tottime = dtime.tv_sec + (dtime.tv_usec / 1e6);
     TEST_info("rcu_torture_result is %d\n", rcu_torture_result);
     TEST_info("performed %d reads and %d writes over 2 read and 2 write threads in %e seconds",
-              reader1_iterations + reader2_iterations,
-              writer1_iterations + writer2_iterations, tottime);
+        reader1_iterations + reader2_iterations,
+        writer1_iterations + writer2_iterations, tottime);
+    if ((reader1_iterations + reader2_iterations == 0)
+        || (writer1_iterations + writer2_iterations == 0)) {
+        TEST_info("Threads did not iterate\n");
+        goto out;
+    }
     avr = tottime / (reader1_iterations + reader2_iterations);
     avw = tottime / (writer1_iterations + writer2_iterations);
     TEST_info("Average read time %e/read", avr);
     TEST_info("Average write time %e/write", avw);
 
+    if (!TEST_int_eq(rcu_torture_result, 1))
+        goto out;
+
+    rc = 1;
+out:
     ossl_rcu_lock_free(rcu_lock);
+    CRYPTO_THREAD_lock_free(atomiclock);
     if (!TEST_int_eq(rcu_torture_result, 1))
         return 0;
 
-    return 1;
+    return rc;
 }
 
 static int torture_rcu_low(void)
@@ -460,7 +503,12 @@ static int torture_rcu_high(void)
     contention = 1;
     return _torture_rcu();
 }
-# endif
+
+static int torture_rcu_high2(void)
+{
+    contention = 2;
+    return _torture_rcu();
+}
 #endif
 
 static CRYPTO_ONCE once_run = CRYPTO_ONCE_STATIC_INIT;
@@ -511,7 +559,7 @@ static void thread_local_thread_cb(void)
     ptr = CRYPTO_THREAD_get_local(&thread_local_key);
     if (!TEST_ptr_null(ptr)
         || !TEST_true(CRYPTO_THREAD_set_local(&thread_local_key,
-                                              &destructor_run_count)))
+            &destructor_run_count)))
         return;
 
     ptr = CRYPTO_THREAD_get_local(&thread_local_key);
@@ -527,7 +575,7 @@ static int test_thread_local(void)
     void *ptr = NULL;
 
     if (!TEST_true(CRYPTO_THREAD_init_local(&thread_local_key,
-                                            thread_local_destructor)))
+            thread_local_destructor)))
         return 0;
 
     ptr = CRYPTO_THREAD_get_local(&thread_local_key);
@@ -543,10 +591,10 @@ static int test_thread_local(void)
     if (!TEST_ptr_null(ptr))
         return 0;
 
-# if !defined(OPENSSL_SYS_WINDOWS)
+#if !defined(OPENSSL_SYS_WINDOWS)
     if (!TEST_int_eq(destructor_run_count, 1))
         return 0;
-# endif
+#endif
 #endif
 
     if (!TEST_true(CRYPTO_THREAD_cleanup_local(&thread_local_key)))
@@ -554,6 +602,29 @@ static int test_thread_local(void)
     return 1;
 }
 
+/*
+ * Basic test to ensure that we can repeatedly create and
+ * destroy local keys without leaking anything
+ */
+static int test_thread_local_multi_key(void)
+{
+    int dummy;
+    int i;
+
+    for (i = 0; i < 1000; i++) {
+        if (!TEST_true(CRYPTO_THREAD_init_local(&thread_local_key,
+                thread_local_destructor)))
+            return 0;
+
+        if (!TEST_true(CRYPTO_THREAD_set_local(&thread_local_key, &dummy)))
+            return 0;
+
+        if (!TEST_true(CRYPTO_THREAD_cleanup_local(&thread_local_key)))
+            return 0;
+    }
+    return 1;
+}
+
 static int test_atomic(void)
 {
     int val = 0, ret = 0, testresult = 0;
@@ -583,12 +654,12 @@ static int test_atomic(void)
     if (CRYPTO_atomic_or(&val64, 2, &ret64, NULL)) {
         /* This succeeds therefore we're on a platform with lockless atomics */
         if (!TEST_uint_eq((unsigned int)val64, 3)
-                || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
+            || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
             goto err;
     } else {
         /* This failed therefore we're on a platform without lockless atomics */
         if (!TEST_uint_eq((unsigned int)val64, 1)
-                || !TEST_int_eq((unsigned int)ret64, 0))
+            || !TEST_int_eq((unsigned int)ret64, 0))
             goto err;
     }
     val64 = 1;
@@ -598,19 +669,19 @@ static int test_atomic(void)
         goto err;
 
     if (!TEST_uint_eq((unsigned int)val64, 3)
-            || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
+        || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
         goto err;
 
     ret64 = 0;
     if (CRYPTO_atomic_load(&val64, &ret64, NULL)) {
         /* This succeeds therefore we're on a platform with lockless atomics */
         if (!TEST_uint_eq((unsigned int)val64, 3)
-                || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
+            || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
             goto err;
     } else {
         /* This failed therefore we're on a platform without lockless atomics */
         if (!TEST_uint_eq((unsigned int)val64, 3)
-                || !TEST_int_eq((unsigned int)ret64, 0))
+            || !TEST_int_eq((unsigned int)ret64, 0))
             goto err;
     }
 
@@ -619,11 +690,57 @@ static int test_atomic(void)
         goto err;
 
     if (!TEST_uint_eq((unsigned int)val64, 3)
+        || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
+        goto err;
+
+    ret64 = 0;
+
+    if (CRYPTO_atomic_and(&val64, 5, &ret64, NULL)) {
+        /* This succeeds therefore we're on a platform with lockless atomics */
+        if (!TEST_uint_eq((unsigned int)val64, 1)
             || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
+            goto err;
+    } else {
+        /* This failed therefore we're on a platform without lockless atomics */
+        if (!TEST_uint_eq((unsigned int)val64, 3)
+            || !TEST_int_eq((unsigned int)ret64, 0))
+            goto err;
+    }
+    val64 = 3;
+    ret64 = 0;
+
+    if (!TEST_true(CRYPTO_atomic_and(&val64, 5, &ret64, lock)))
+        goto err;
+
+    if (!TEST_uint_eq((unsigned int)val64, 1)
+        || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
+        goto err;
+
+    ret64 = 0;
+
+    if (CRYPTO_atomic_add64(&val64, 2, &ret64, NULL)) {
+        /* This succeeds therefore we're on a platform with lockless atomics */
+        if (!TEST_uint_eq((unsigned int)val64, 3)
+            || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
+            goto err;
+    } else {
+        /* This failed therefore we're on a platform without lockless atomics */
+        if (!TEST_uint_eq((unsigned int)val64, 1)
+            || !TEST_int_eq((unsigned int)ret64, 0))
+            goto err;
+    }
+    val64 = 1;
+    ret64 = 0;
+
+    if (!TEST_true(CRYPTO_atomic_add64(&val64, 2, &ret64, lock)))
+        goto err;
+
+    if (!TEST_uint_eq((unsigned int)val64, 3)
+        || !TEST_uint_eq((unsigned int)val64, (unsigned int)ret64))
         goto err;
 
     testresult = 1;
- err:
+err:
     CRYPTO_THREAD_lock_free(lock);
     return testresult;
 }
@@ -670,15 +787,14 @@ static int thread_setup_libctx(int libctx, const char *providers[])
 {
     size_t n;
 
-    if (libctx && !TEST_true(test_get_libctx(&multi_libctx, NULL, config_file,
-                                             NULL, NULL)))
+    if (libctx && !TEST_true(test_get_libctx(&multi_libctx, NULL, config_file, NULL, NULL)))
         return 0;
 
     if (providers != NULL)
         for (n = 0; providers[n] != NULL; n++)
             if (!TEST_size_t_lt(n, MAXIMUM_PROVIDERS)
                 || !TEST_ptr(multi_provider[n] = OSSL_PROVIDER_load(multi_libctx,
-                                                                    providers[n]))) {
+                                 providers[n]))) {
                 thead_teardown_libctx();
                 return 0;
             }
@@ -702,7 +818,7 @@ static int start_threads(size_t n, void (*thread_func)(void))
     if (!TEST_size_t_le(multi_num_threads + n, MAXIMUM_THREADS))
         return 0;
 
-    for (i = 0 ; i < n; i++)
+    for (i = 0; i < n; i++)
         if (!TEST_true(run_thread(multi_threads + multi_num_threads++, thread_func)))
             return 0;
     return 1;
@@ -710,24 +826,24 @@ static int start_threads(size_t n, void (*thread_func)(void))
 
 /* Template multi-threaded test function */
 static int thread_run_test(void (*main_func)(void),
-                           size_t num_threads, void (*thread_func)(void),
-                           int libctx, const char *providers[])
+    size_t num_threads, void (*thread_func)(void),
+    int libctx, const char *providers[])
 {
     int testresult = 0;
 
     multi_intialise();
     if (!thread_setup_libctx(libctx, providers)
-            || !start_threads(num_threads, thread_func))
+        || !start_threads(num_threads, thread_func))
         goto err;
 
     if (main_func != NULL)
         main_func();
 
     if (!teardown_threads()
-            || !TEST_true(multi_success))
+        || !TEST_true(multi_success))
         goto err;
     testresult = 1;
- err:
+err:
     thead_teardown_libctx();
     return testresult;
 }
@@ -759,24 +875,24 @@ static void thread_general_worker(void)
     isfips = OSSL_PROVIDER_available(multi_libctx, "fips");
 
     if (!TEST_ptr(mdctx)
-            || !TEST_ptr(md)
-            || !TEST_ptr(cipherctx)
-            || !TEST_ptr(ciph))
+        || !TEST_ptr(md)
+        || !TEST_ptr(cipherctx)
+        || !TEST_ptr(ciph))
         goto err;
 
     /* Do some work */
     for (i = 0; i < 5; i++) {
         if (!TEST_true(EVP_DigestInit_ex(mdctx, md, NULL))
-                || !TEST_true(EVP_DigestUpdate(mdctx, message, messlen))
-                || !TEST_true(EVP_DigestFinal(mdctx, out, &mdoutl)))
+            || !TEST_true(EVP_DigestUpdate(mdctx, message, messlen))
+            || !TEST_true(EVP_DigestFinal(mdctx, out, &mdoutl)))
             goto err;
     }
     for (i = 0; i < 5; i++) {
         if (!TEST_true(EVP_EncryptInit_ex(cipherctx, ciph, NULL, key, iv))
-                || !TEST_true(EVP_EncryptUpdate(cipherctx, out, &ciphoutl,
-                                                (unsigned char *)message,
-                                                messlen))
-                || !TEST_true(EVP_EncryptFinal(cipherctx, out, &ciphoutl)))
+            || !TEST_true(EVP_EncryptUpdate(cipherctx, out, &ciphoutl,
+                (unsigned char *)message,
+                (int)messlen))
+            || !TEST_true(EVP_EncryptFinal(cipherctx, out, &ciphoutl)))
             goto err;
     }
 
@@ -785,12 +901,12 @@ static void thread_general_worker(void)
      * Therefore we use an insecure bit length where we can (512).
      * In the FIPS module though we must use a longer length.
      */
-    pkey = EVP_PKEY_Q_keygen(multi_libctx, NULL, "RSA", isfips ? 2048 : 512);
+    pkey = EVP_PKEY_Q_keygen(multi_libctx, NULL, "RSA", (size_t)(isfips ? 2048 : 512));
     if (!TEST_ptr(pkey))
         goto err;
 
     testresult = 1;
- err:
+err:
     EVP_MD_CTX_free(mdctx);
     EVP_MD_free(md);
     EVP_CIPHER_CTX_free(cipherctx);
@@ -826,15 +942,15 @@ static void thread_shared_evp_pkey(void)
         if (i > 0)
             EVP_PKEY_CTX_free(ctx);
         ctx = EVP_PKEY_CTX_new_from_pkey(multi_libctx, shared_evp_pkey,
-                                         i == 0 ? "provider=default"
-                                                : "provider=fips");
+            i == 0 ? "provider=default"
+                   : "provider=fips");
         if (!TEST_ptr(ctx))
             goto err;
 
         if (!TEST_int_ge(EVP_PKEY_encrypt_init(ctx), 0)
-                || !TEST_int_ge(EVP_PKEY_encrypt(ctx, ctbuf, &ctlen,
-                                                (unsigned char *)msg, strlen(msg)),
-                                                0))
+            || !TEST_int_ge(EVP_PKEY_encrypt(ctx, ctbuf, &ctlen,
+                                (unsigned char *)msg, strlen(msg)),
+                0))
             goto err;
 
         EVP_PKEY_CTX_free(ctx);
@@ -845,15 +961,15 @@ static void thread_shared_evp_pkey(void)
 
         ptlen = sizeof(ptbuf);
         if (!TEST_int_ge(EVP_PKEY_decrypt_init(ctx), 0)
-                || !TEST_int_gt(EVP_PKEY_decrypt(ctx, ptbuf, &ptlen, ctbuf, ctlen),
-                                                0)
-                || !TEST_mem_eq(msg, strlen(msg), ptbuf, ptlen))
+            || !TEST_int_gt(EVP_PKEY_decrypt(ctx, ptbuf, &ptlen, ctbuf, ctlen),
+                0)
+            || !TEST_mem_eq(msg, strlen(msg), ptbuf, ptlen))
             goto err;
     }
 
     success = 1;
 
- err:
+err:
     EVP_PKEY_CTX_free(ctx);
     if (!success)
         multi_set_success(0);
@@ -864,7 +980,7 @@ static void thread_provider_load_unload(void)
     OSSL_PROVIDER *deflt = OSSL_PROVIDER_load(multi_libctx, "default");
 
     if (!TEST_ptr(deflt)
-            || !TEST_true(OSSL_PROVIDER_available(multi_libctx, "default")))
+        || !TEST_true(OSSL_PROVIDER_available(multi_libctx, "default")))
         multi_set_success(0);
 
     OSSL_PROVIDER_unload(deflt);
@@ -873,7 +989,7 @@ static void thread_provider_load_unload(void)
 static int test_multi_general_worker_default_provider(void)
 {
     return thread_run_test(&thread_general_worker, 2, &thread_general_worker,
-                           1, default_provider);
+        1, default_provider);
 }
 
 static int test_multi_general_worker_fips_provider(void)
@@ -881,13 +997,13 @@ static int test_multi_general_worker_fips_provider(void)
     if (!do_fips)
         return TEST_skip("FIPS not supported");
     return thread_run_test(&thread_general_worker, 2, &thread_general_worker,
-                           1, fips_provider);
+        1, fips_provider);
 }
 
 static int test_multi_fetch_worker(void)
 {
     return thread_run_test(&thread_multi_simple_fetch,
-                           2, &thread_multi_simple_fetch, 1, default_provider);
+        2, &thread_multi_simple_fetch, 1, default_provider);
 }
 
 static int test_multi_shared_pkey_common(void (*worker)(void))
@@ -895,20 +1011,19 @@ static int test_multi_shared_pkey_common(void (*worker)(void))
     int testresult = 0;
 
     multi_intialise();
-    if (!thread_setup_libctx(1, do_fips ? fips_and_default_providers
-                                        : default_provider)
-            || !TEST_ptr(shared_evp_pkey = load_pkey_pem(privkey, multi_libctx))
-            || !start_threads(1, &thread_shared_evp_pkey)
-            || !start_threads(1, worker))
+    if (!thread_setup_libctx(1, do_fips ? fips_and_default_providers : default_provider)
+        || !TEST_ptr(shared_evp_pkey = load_pkey_pem(privkey, multi_libctx))
+        || !start_threads(1, &thread_shared_evp_pkey)
+        || !start_threads(1, worker))
         goto err;
 
     thread_shared_evp_pkey();
 
     if (!teardown_threads()
-            || !TEST_true(multi_success))
+        || !TEST_true(multi_success))
         goto err;
     testresult = 1;
- err:
+err:
     EVP_PKEY_free(shared_evp_pkey);
     thead_teardown_libctx();
     return testresult;
@@ -936,6 +1051,44 @@ static int test_multi_shared_pkey(void)
     return test_multi_shared_pkey_common(&thread_shared_evp_pkey);
 }
 
+static void thread_release_shared_pkey(void)
+{
+    OSSL_sleep(0);
+    EVP_PKEY_free(shared_evp_pkey);
+}
+
+static int test_multi_shared_pkey_release(void)
+{
+    int testresult = 0;
+    size_t i = 1;
+
+    multi_intialise();
+    shared_evp_pkey = NULL;
+    if (!thread_setup_libctx(1, do_fips ? fips_and_default_providers : default_provider)
+        || !TEST_ptr(shared_evp_pkey = load_pkey_pem(privkey, multi_libctx)))
+        goto err;
+    for (; i < 10; ++i) {
+        if (!TEST_true(EVP_PKEY_up_ref(shared_evp_pkey)))
+            goto err;
+    }
+
+    if (!start_threads(10, &thread_release_shared_pkey))
+        goto err;
+    i = 0;
+
+    if (!teardown_threads()
+        || !TEST_true(multi_success))
+        goto err;
+    testresult = 1;
+err:
+    while (i > 0) {
+        EVP_PKEY_free(shared_evp_pkey);
+        --i;
+    }
+    thead_teardown_libctx();
+    return testresult;
+}
+
 static int test_multi_load_unload_provider(void)
 {
     EVP_MD *sha256 = NULL;
@@ -944,9 +1097,9 @@ static int test_multi_load_unload_provider(void)
 
     multi_intialise();
     if (!thread_setup_libctx(1, NULL)
-            || !TEST_ptr(prov = OSSL_PROVIDER_load(multi_libctx, "default"))
-            || !TEST_ptr(sha256 = EVP_MD_fetch(multi_libctx, "SHA2-256", NULL))
-            || !TEST_true(OSSL_PROVIDER_unload(prov)))
+        || !TEST_ptr(prov = OSSL_PROVIDER_load(multi_libctx, "default"))
+        || !TEST_ptr(sha256 = EVP_MD_fetch(multi_libctx, "SHA2-256", NULL))
+        || !TEST_true(OSSL_PROVIDER_unload(prov)))
         goto err;
     prov = NULL;
 
@@ -956,10 +1109,10 @@ static int test_multi_load_unload_provider(void)
     thread_provider_load_unload();
 
     if (!teardown_threads()
-            || !TEST_true(multi_success))
+        || !TEST_true(multi_success))
         goto err;
     testresult = 1;
- err:
+err:
     OSSL_PROVIDER_unload(prov);
     EVP_MD_free(sha256);
     thead_teardown_libctx();
@@ -977,7 +1130,7 @@ static void test_multi_load_worker(void)
     OSSL_PROVIDER *prov;
 
     if (!TEST_ptr(prov = OSSL_PROVIDER_load(multi_libctx, multi_load_provider))
-            || !TEST_true(OSSL_PROVIDER_unload(prov)))
+        || !TEST_true(OSSL_PROVIDER_unload(prov)))
         multi_set_success(0);
 }
 
@@ -991,7 +1144,7 @@ static int test_multi_default(void)
     multidefault_run = 1;
 
     return thread_run_test(&thread_multi_simple_fetch,
-                           2, &thread_multi_simple_fetch, 0, default_provider);
+        2, &thread_multi_simple_fetch, 0, NULL);
 }
 
 static int test_multi_load(void)
@@ -1019,7 +1172,8 @@ static int test_multi_load(void)
     OSSL_PROVIDER_unload(prov);
 
     return thread_run_test(NULL, MAXIMUM_THREADS, &test_multi_load_worker, 0,
-                          NULL) && res;
+               NULL)
+        && res;
 }
 
 static void test_obj_create_one(void)
@@ -1032,29 +1186,16 @@ static void test_obj_create_one(void)
     BIO_snprintf(sn, sizeof(sn), "short-name-%s", tids);
     BIO_snprintf(ln, sizeof(ln), "long-name-%s", tids);
     if (!TEST_int_ne(id, 0)
-            || !TEST_true(id = OBJ_create(oid, sn, ln))
-            || !TEST_true(OBJ_add_sigid(id, NID_sha3_256, NID_rsa)))
+        || !TEST_true(id = OBJ_create(oid, sn, ln))
+        || !TEST_true(OBJ_add_sigid(id, NID_sha3_256, NID_rsa)))
         multi_set_success(0);
 }
 
 static int test_obj_add(void)
 {
     return thread_run_test(&test_obj_create_one,
-                           MAXIMUM_THREADS, &test_obj_create_one,
-                           1, default_provider);
-}
-
-static void test_lib_ctx_load_config_worker(void)
-{
-    if (!TEST_int_eq(OSSL_LIB_CTX_load_config(multi_libctx, config_file), 1))
-        multi_set_success(0);
-}
-
-static int test_lib_ctx_load_config(void)
-{
-    return thread_run_test(&test_lib_ctx_load_config_worker,
-                           MAXIMUM_THREADS, &test_lib_ctx_load_config_worker,
-                           1, default_provider);
+        MAXIMUM_THREADS, &test_obj_create_one,
+        1, default_provider);
 }
 
 #if !defined(OPENSSL_NO_DGRAM) && !defined(OPENSSL_NO_SOCK)
@@ -1066,13 +1207,13 @@ static void test_bio_dgram_pair_worker(void)
     int ok = 0;
     uint8_t ch = 0;
     uint8_t scratch[64];
-    BIO_MSG msg = {0};
+    BIO_MSG msg = { 0 };
     size_t num_processed = 0;
 
     if (!TEST_int_eq(RAND_bytes_ex(multi_libctx, &ch, 1, 64), 1))
         goto err;
 
-    msg.data     = scratch;
+    msg.data = scratch;
     msg.data_len = sizeof(scratch);
 
     /*
@@ -1081,10 +1222,10 @@ static void test_bio_dgram_pair_worker(void)
      */
     if (ch & 2)
         r = BIO_sendmmsg(ch & 1 ? multi_bio2 : multi_bio1, &msg,
-                         sizeof(BIO_MSG), 1, 0, &num_processed);
+            sizeof(BIO_MSG), 1, 0, &num_processed);
     else
         r = BIO_recvmmsg(ch & 1 ? multi_bio2 : multi_bio1, &msg,
-                         sizeof(BIO_MSG), 1, 0, &num_processed);
+            sizeof(BIO_MSG), 1, 0, &num_processed);
 
     ok = 1;
 err:
@@ -1104,9 +1245,9 @@ static int test_bio_dgram_pair(void)
     multi_bio1 = bio1;
     multi_bio2 = bio2;
 
-    r  = thread_run_test(&test_bio_dgram_pair_worker,
-                         MAXIMUM_THREADS, &test_bio_dgram_pair_worker,
-                         1, default_provider);
+    r = thread_run_test(&test_bio_dgram_pair_worker,
+        MAXIMUM_THREADS, &test_bio_dgram_pair_worker,
+        1, default_provider);
 
 err:
     BIO_free(bio1);
@@ -1141,7 +1282,7 @@ static void test_pem_read_one(void)
         goto err;
     }
 
-    pem = BIO_new_mem_buf(pemdata, len);
+    pem = BIO_new_mem_buf(pemdata, (int)len);
     if (pem == NULL) {
         multi_set_success(0);
         goto err;
@@ -1151,7 +1292,7 @@ static void test_pem_read_one(void)
     if (key == NULL)
         multi_set_success(0);
 
- err:
+err:
     EVP_PKEY_free(key);
     BIO_free(pem);
     OPENSSL_free(pemdata);
@@ -1161,13 +1302,106 @@ static void test_pem_read_one(void)
 static int test_pem_read(void)
 {
     return thread_run_test(&test_pem_read_one, MAXIMUM_THREADS,
-                           &test_pem_read_one, 1, default_provider);
+        &test_pem_read_one, 1, default_provider);
+}
+
+static X509_STORE *store = NULL;
+
+static void test_x509_store_by_subject(void)
+{
+    X509_STORE_CTX *ctx;
+    X509_OBJECT *obj = NULL;
+    X509_NAME *name = NULL;
+    int success = 0;
+
+    ctx = X509_STORE_CTX_new();
+    if (!TEST_ptr(ctx))
+        goto err;
+
+    if (!TEST_true(X509_STORE_CTX_init(ctx, store, NULL, NULL)))
+        goto err;
+
+    name = X509_NAME_new();
+    if (!TEST_ptr(name))
+        goto err;
+    if (!TEST_true(X509_NAME_add_entry_by_txt(name, "CN", MBSTRING_ASC,
+            (unsigned char *)"Root CA",
+            -1, -1, 0)))
+        goto err;
+    obj = X509_STORE_CTX_get_obj_by_subject(ctx, X509_LU_X509, name);
+    if (!TEST_ptr(obj))
+        goto err;
+
+    success = 1;
+err:
+    X509_OBJECT_free(obj);
+    X509_STORE_CTX_free(ctx);
+    X509_NAME_free(name);
+    if (!success)
+        multi_set_success(0);
+}
+
+/* Test accessing an X509_STORE from multiple threads */
+static int test_x509_store(void)
+{
+    int ret = 0;
+
+    store = X509_STORE_new();
+    if (!TEST_ptr(store))
+        return 0;
+    if (!TEST_true(X509_STORE_load_store(store, storedir)))
+        goto err;
+
+    ret = thread_run_test(&test_x509_store_by_subject, MAXIMUM_THREADS,
+        &test_x509_store_by_subject, 0, NULL);
+
+err:
+    X509_STORE_free(store);
+    store = NULL;
+    return ret;
+}
+
+/* Test using OBJ_create in multiple threads */
+static void test_obj_create_worker(void)
+{
+    int i, nid, nid2;
+    time_t now;
+    char name[40];
+
+    for (i = 0; i < 4; i++) {
+        now = time(NULL);
+        snprintf(name, sizeof(name), "Time in Seconds = %ld", (long)now);
+        while (now == time(NULL))
+            /* no-op */;
+        nid = OBJ_create(NULL, NULL, name);
+        nid2 = OBJ_ln2nid(name);
+        if (nid != NID_undef) {
+            if (nid2 != nid) {
+                TEST_info("oops: name='%s' nid=%d nid2=%d", name, nid, nid2);
+                multi_set_success(0);
+                break;
+            }
+        } else {
+            if (nid2 == NID_undef) {
+                TEST_info("oops: name='%s' nid=%d nid2=%d", name, nid, nid2);
+                multi_set_success(0);
+                break;
+            }
+        }
+    }
+}
+
+static int test_obj_stress(void)
+{
+    return thread_run_test(&test_obj_create_worker, MAXIMUM_THREADS,
+        &test_obj_create_worker, 0, NULL);
 }
 
 typedef enum OPTION_choice {
     OPT_ERR = -1,
     OPT_EOF = 0,
-    OPT_FIPS, OPT_CONFIG_FILE,
+    OPT_FIPS,
+    OPT_CONFIG_FILE,
     OPT_TEST_ENUM
 } OPTION_CHOICE;
 
@@ -1177,7 +1411,7 @@ const OPTIONS *test_get_options(void)
         OPT_TEST_OPTIONS_DEFAULT_USAGE,
         { "fips", OPT_FIPS, '-', "Test the FIPS provider" },
         { "config", OPT_CONFIG_FILE, '<',
-          "The configuration file to use for the libctx" },
+            "The configuration file to use for the libctx" },
         { NULL }
     };
     return options;
@@ -1210,6 +1444,10 @@ int setup_tests(void)
     if (!TEST_ptr(privkey))
         return 0;
 
+    storedir = test_mk_file_path(datadir, "store");
+    if (!TEST_ptr(storedir))
+        return 0;
+
     if (!TEST_ptr(global_lock = CRYPTO_THREAD_lock_new()))
         return 0;
 
@@ -1225,13 +1463,13 @@ int setup_tests(void)
 #if defined(OPENSSL_THREADS)
     ADD_TEST(torture_rw_low);
     ADD_TEST(torture_rw_high);
-# ifndef OPENSSL_SYS_MACOSX
     ADD_TEST(torture_rcu_low);
     ADD_TEST(torture_rcu_high);
-# endif
+    ADD_TEST(torture_rcu_high2);
 #endif
     ADD_TEST(test_once);
     ADD_TEST(test_thread_local);
+    ADD_TEST(test_thread_local_multi_key);
     ADD_TEST(test_atomic);
     ADD_TEST(test_multi_load);
     ADD_TEST(test_multi_general_worker_default_provider);
@@ -1241,19 +1479,22 @@ int setup_tests(void)
 #ifndef OPENSSL_NO_DEPRECATED_3_0
     ADD_TEST(test_multi_downgrade_shared_pkey);
 #endif
+    ADD_TEST(test_multi_shared_pkey_release);
     ADD_TEST(test_multi_load_unload_provider);
     ADD_TEST(test_obj_add);
-    ADD_TEST(test_lib_ctx_load_config);
 #if !defined(OPENSSL_NO_DGRAM) && !defined(OPENSSL_NO_SOCK)
     ADD_TEST(test_bio_dgram_pair);
 #endif
     ADD_TEST(test_pem_read);
+    ADD_TEST(test_x509_store);
+    ADD_TEST(test_obj_stress);
     return 1;
 }
 
 void cleanup_tests(void)
 {
     OPENSSL_free(privkey);
+    OPENSSL_free(storedir);
 #ifdef TSAN_REQUIRES_LOCKING
     CRYPTO_THREAD_lock_free(tsan_lock);
 #endif
